# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <github@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
from collections import OrderedDict, defaultdict
from unittest import TestCase

from df_config.config.dynamic_settings import RawValue, SettingReference
from df_config.config.merger import SettingMerger
from df_config.config.values_providers import DictProvider


class TestSettingMerger(TestCase):
    def test_priority(self):
        merger = SettingMerger(
            None,
            [DictProvider({"X": 1}, name="d1"), DictProvider({"X": 2}, name="d2")],
        )
        merger.process()
        merger.post_process()
        self.assertEqual({"X": 2}, merger.settings)
        self.assertEqual(OrderedDict([("d1", 1), ("d2", 2)]), merger.raw_settings["X"])

    def test_postprocess(self):
        merger = SettingMerger(
            None,
            [
                DictProvider(
                    {
                        "INSTALLED_APPS": [
                            "df_config",
                            "django.contrib.auth",
                            "df_config",
                        ]
                    },
                    name="d1",
                )
            ],
        )
        merger.process()
        self.assertEqual(
            {"INSTALLED_APPS": ["df_config", "django.contrib.auth", "df_config"]},
            merger.settings,
        )
        merger.post_process()
        self.assertEqual(
            {"INSTALLED_APPS": ["df_config", "django.contrib.auth"]}, merger.settings
        )

    def test_parse(self):
        merger = SettingMerger(
            None,
            [
                DictProvider({"X": 1, "Y": "x{X}"}, name="1"),
                DictProvider({"X": 2}, name="2"),
            ],
        )
        merger.process()
        self.assertEqual({"X": 2, "Y": "x2",}, merger.settings)

    def test_loop(self):
        merger = SettingMerger(
            None, [DictProvider({"X": "{Y}", "Y": "{Z}", "Z": "{X}"}, name="1")],
        )
        self.assertRaises(ValueError, merger.process)

    def test_dynamic_setting(self):
        merger = SettingMerger(
            None,
            [
                DictProvider({"X": 1}, name="d1"),
                DictProvider({"X": RawValue("{Y}")}, name="d2"),
            ],
        )
        merger.process()
        self.assertEqual({"X": "{Y}"}, merger.settings)
        self.assertEqual(
            OrderedDict([("d1", 1), ("d2", RawValue("{Y}"))]), merger.raw_settings["X"]
        )

    def test_complex_settings_ref(self):
        merger = SettingMerger(
            None,
            [
                DictProvider({"X": 1}, name="d1"),
                DictProvider({"Y": SettingReference("X")}, name="d2"),
            ],
        )
        merger.process()
        self.assertEqual({"X": 1, "Y": 1}, merger.settings)

    def test_complex_settings_ref_str(self):
        merger = SettingMerger(
            None,
            [DictProvider({"X": 1}, name="d1"), DictProvider({"Y": "{X}"}, name="d2")],
        )
        merger.process()
        self.assertEqual({"X": 1, "Y": "1"}, merger.settings)

    def test_complex_settings_list(self):
        merger = SettingMerger(
            None,
            [
                DictProvider({"X": 1}, name="d1"),
                DictProvider({"Y": ["{X}"]}, name="d2"),
            ],
        )
        merger.process()
        self.assertEqual({"X": 1, "Y": ["1"]}, merger.settings)

    def test_complex_settings_tuple(self):
        merger = SettingMerger(
            None,
            [
                DictProvider({"X": 1}, name="d1"),
                DictProvider({"Y": ("{X}",)}, name="d2"),
            ],
        )
        merger.process()
        self.assertEqual({"X": 1, "Y": ("1",)}, merger.settings)

    def test_complex_settings_set(self):
        merger = SettingMerger(
            None,
            [
                DictProvider({"X": 1}, name="d1"),
                DictProvider({"Y": {"{X}"}}, name="d2"),
            ],
        )
        merger.process()
        self.assertEqual({"X": 1, "Y": {"1"}}, merger.settings)

    def test_complex_settings_dict(self):
        merger = SettingMerger(
            None,
            [
                DictProvider({"X": 1}, name="d1"),
                DictProvider({"Y": {"{X}": "{X}"}}, name="d2"),
            ],
        )
        merger.process()
        self.assertEqual({"X": 1, "Y": {"1": "1"}}, merger.settings)

    def test_complex_settings_ordereddict(self):
        merger = SettingMerger(
            None,
            [
                DictProvider({"X": 1}, name="d1"),
                DictProvider({"Y": OrderedDict([("{X}", "{X}")])}, name="d2"),
            ],
        )
        merger.process()
        self.assertEqual({"X": 1, "Y": {"1": "1"}}, merger.settings)
        self.assertIsInstance(merger.settings["Y"], OrderedDict)

    def test_complex_settings_defaultdict(self):
        values = defaultdict(lambda: [])
        values["{X}"].append("{X}")
        merger = SettingMerger(
            None,
            [DictProvider({"X": 1}, name="d1"), DictProvider({"Y": values}, name="d2")],
        )
        merger.process()
        self.assertEqual({"X": 1, "Y": {"1": ["1"]}}, merger.settings)
        self.assertIsInstance(merger.settings["Y"], defaultdict)
        merger.settings["Y"]["2"].append(1)



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <github@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################


# noinspection PyMethodMayBeStatic
import grp
import os
import pwd

from django.core.checks import Error
from django.utils.module_loading import import_string
from pkg_resources import DistributionNotFound, get_distribution

from df_config.checks import missing_package, settings_check_results


# noinspection PyMethodMayBeStatic
class AuthenticationBackends:
    required_settings = [
        "ALLAUTH_PROVIDER_APPS",
        "DF_REMOTE_USER_HEADER",
        "AUTH_LDAP_SERVER_URI",
        "USE_PAM_AUTHENTICATION",
        "DF_ALLOW_LOCAL_USERS",
        "USE_ALL_AUTH",
        "RADIUS_SERVER",
    ]

    def __call__(self, settings_dict):
        backends = []
        backends += self.process_remote_user(settings_dict)
        backends += self.process_radius(settings_dict)
        backends += self.process_django(settings_dict)
        backends += self.process_django_ldap(settings_dict)
        backends += self.process_allauth(settings_dict)
        backends += self.process_pam(settings_dict)
        return backends

    def process_django(self, settings_dict):
        if settings_dict["DF_ALLOW_LOCAL_USERS"]:
            return ["django.contrib.auth.backends.ModelBackend"]
        return []

    def process_remote_user(self, settings_dict):
        if settings_dict["DF_REMOTE_USER_HEADER"]:
            return ["df_config.apps.backends.DefaultGroupsRemoteUserBackend"]
        return []

    def process_allauth(self, settings_dict):
        if (
            not settings_dict["USE_ALL_AUTH"]
            and not settings_dict["ALLAUTH_PROVIDER_APPS"]
        ):
            return []
        try:
            get_distribution("django-allauth")
            return ["allauth.account.auth_backends.AuthenticationBackend"]
        except DistributionNotFound:
            return []

    def process_radius(self, settings_dict):
        if not settings_dict["RADIUS_SERVER"]:
            return []
        try:
            get_distribution("django-radius")
        except DistributionNotFound:
            settings_check_results.append(
                missing_package("django-radius", " to use RADIUS authentication")
            )
            return []
        return ["radiusauth.backends.RADIUSBackend"]

    def process_django_ldap(self, settings_dict):
        if not settings_dict["AUTH_LDAP_SERVER_URI"]:
            return []
        try:
            get_distribution("django-auth-ldap")
        except DistributionNotFound:
            settings_check_results.append(
                missing_package("django-auth-ldap", " to use LDAP authentication")
            )
            return []
        return ["django_auth_ldap.backend.LDAPBackend"]

    def process_pam(self, settings_dict):
        if not settings_dict["USE_PAM_AUTHENTICATION"]:
            return []
        try:
            get_distribution("django_pam")
        except DistributionNotFound:
            settings_check_results.append(
                missing_package("django-pam", " to use PAM authentication")
            )
            return []
        # check if the current user is in the shadow group
        username = pwd.getpwuid(os.getuid()).pw_name
        if not any(
            x.gr_name == "shadow" and username in x.gr_mem for x in grp.getgrall()
        ):
            settings_check_results.append(
                Error(
                    'The user "%s" must belong to the "shadow" group to use PAM '
                    "authentication." % username,
                    obj="configuration",
                )
            )
            return []
        return ["django_pam.auth.backends.PAMBackend"]

    def __repr__(self):
        return "%s.%s" % (self.__module__, "authentication_backends")


authentication_backends = AuthenticationBackends()


def ldap_user_search(settings_dict):
    if (
        settings_dict["AUTH_LDAP_SERVER_URI"]
        and settings_dict["AUTH_LDAP_USER_SEARCH_BASE"]
    ):
        try:
            # noinspection PyPackageRequirements,PyUnresolvedReferences
            import ldap

            # noinspection PyUnresolvedReferences
            from django_auth_ldap.config import LDAPSearch
        except ImportError:
            return None
        return LDAPSearch(
            settings_dict["AUTH_LDAP_USER_SEARCH_BASE"],
            ldap.SCOPE_SUBTREE,
            settings_dict["AUTH_LDAP_FILTER"],
        )
    return None


ldap_user_search.required_settings = [
    "AUTH_LDAP_USER_SEARCH_BASE",
    "AUTH_LDAP_SERVER_URI",
    "AUTH_LDAP_FILTER",
]


def ldap_group_search(settings_dict):
    if (
        settings_dict["AUTH_LDAP_SERVER_URI"]
        and settings_dict["AUTH_LDAP_GROUP_SEARCH_BASE"]
    ):
        try:
            # noinspection PyPackageRequirements,PyUnresolvedReferences
            import ldap

            # noinspection PyUnresolvedReferences
            from django_auth_ldap.config import LDAPSearch
        except ImportError:
            return None
        return LDAPSearch(
            settings_dict["AUTH_LDAP_GROUP_SEARCH_BASE"],
            ldap.SCOPE_SUBTREE,
            "(objectClass=*)",
        )
    return None


ldap_group_search.required_settings = [
    "AUTH_LDAP_GROUP_SEARCH_BASE",
    "AUTH_LDAP_SERVER_URI",
]


def ldap_attribute_map(settings_dict):
    result = {}
    if settings_dict["AUTH_LDAP_USER_FIRST_NAME"]:
        result["first_name"] = settings_dict["AUTH_LDAP_USER_FIRST_NAME"]
    if settings_dict["AUTH_LDAP_USER_LAST_NAME"]:
        result["last_name"] = settings_dict["AUTH_LDAP_USER_LAST_NAME"]
    if settings_dict["AUTH_LDAP_USER_EMAIL"]:
        result["email"] = settings_dict["AUTH_LDAP_USER_EMAIL"]
    return result


ldap_attribute_map.required_settings = [
    "AUTH_LDAP_USER_FIRST_NAME",
    "AUTH_LDAP_USER_LAST_NAME",
    "AUTH_LDAP_USER_EMAIL",
]


def ldap_boolean_attribute_map(settings_dict):
    result = {}
    if settings_dict["AUTH_LDAP_USER_IS_ACTIVE"]:
        result["is_active"] = settings_dict["AUTH_LDAP_USER_IS_ACTIVE"]
    if settings_dict["AUTH_LDAP_USER_IS_STAFF"]:
        result["is_staff"] = settings_dict["AUTH_LDAP_USER_IS_STAFF"]
    if settings_dict["AUTH_LDAP_USER_IS_ACTIVE"]:
        result["is_superuser"] = settings_dict["AUTH_LDAP_USER_IS_SUPERUSER"]
    return result


ldap_boolean_attribute_map.required_settings = [
    "AUTH_LDAP_USER_IS_ACTIVE",
    "AUTH_LDAP_USER_IS_STAFF",
    "AUTH_LDAP_USER_IS_SUPERUSER",
]


def ldap_group_class(settings_dict):
    if settings_dict["AUTH_LDAP_SERVER_URI"]:
        try:
            cls = import_string(settings_dict["AUTH_LDAP_GROUP_NAME"])
            return cls()
        except ImportError:
            return None
    return None


ldap_group_class.required_settings = ["AUTH_LDAP_GROUP_NAME", "AUTH_LDAP_SERVER_URI"]



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <github@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
import os
import re
from urllib.parse import urlparse

from django.utils.crypto import get_random_string
from pkg_resources import DistributionNotFound, VersionConflict, get_distribution

from df_config.checks import missing_package, settings_check_results
from df_config.config.dynamic_settings import DynamicSettting


def smart_hostname(settings_dict):
    """
    By default, use the listen address and port as server name.
    Use the "HEROKU_APP_NAME" environment variable if present.

    :param settings_dict:
    :return:
    """
    if "HEROKU_APP_NAME" in os.environ:
        return "https://%s.herokuapp.com/" % os.environ["HEROKU_APP_NAME"]
    return "http://%s/" % settings_dict["LISTEN_ADDRESS"]


smart_hostname.required_settings = ["LISTEN_ADDRESS"]


class DefaultListenAddress(DynamicSettting):
    def get_value(self, merger, provider_name: str, setting_name: str):
        port = os.environ.get("PORT", "")
        if re.match(r"^([1-9]\d*)$", port) and 1 <= int(port) <= 65535:
            return "0.0.0.0:%s" % port
        return "localhost:%d" % self.value


def template_setting(settings_dict):
    loaders = [
        "django.template.loaders.filesystem.Loader",
        "django.template.loaders.app_directories.Loader",
    ]
    if settings_dict["DEBUG"]:
        backend = {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "NAME": "default",
            "DIRS": settings_dict["TEMPLATE_DIRS"],
            "OPTIONS": {
                "context_processors": settings_dict["TEMPLATE_CONTEXT_PROCESSORS"],
                "loaders": loaders,
                "debug": True,
            },
        }
    else:
        backend = {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "NAME": "default",
            "DIRS": settings_dict["TEMPLATE_DIRS"],
            "OPTIONS": {
                "context_processors": settings_dict["TEMPLATE_CONTEXT_PROCESSORS"],
                "debug": False,
                "loaders": [("django.template.loaders.cached.Loader", loaders)],
            },
        }
    return [backend]


template_setting.required_settings = [
    "DEBUG",
    "TEMPLATE_DIRS",
    "TEMPLATE_CONTEXT_PROCESSORS",
]


def allowed_hosts(settings_dict):
    result = {"127.0.0.1", "::1", "localhost"}
    listened_ip, sep, port = settings_dict["LISTEN_ADDRESS"].rpartition(":")
    if sep == ":" and listened_ip not in ("::", "0.0.0.0"):
        result.add(listened_ip)
    result.add(settings_dict["SERVER_NAME"])
    return list(sorted(result))


allowed_hosts.required_settings = ["SERVER_NAME", "LISTEN_ADDRESS"]


def secure_hsts_seconds(settings_dict):
    if settings_dict["USE_SSL"]:
        return 86400
    return 0


secure_hsts_seconds.required_settings = ["USE_SSL"]


def url_parse_server_name(settings_dict):
    """Return the public hostname, given the public base URL

    >>> url_parse_server_name({'SERVER_BASE_URL': 'https://demo.example.org/'})
    'demo.example.org'

    """
    return urlparse(settings_dict["SERVER_BASE_URL"]).hostname


url_parse_server_name.required_settings = ["SERVER_BASE_URL"]


def url_parse_server_port(settings_dict):
    """Return the public port, given the public base URL

    >>> url_parse_server_port({'SERVER_BASE_URL': 'https://demo.example.org/', 'USE_SSL': True})
    443
    >>> url_parse_server_port({'SERVER_BASE_URL': 'http://demo.example.org/', 'USE_SSL': False})
    80
    >>> url_parse_server_port({'SERVER_BASE_URL': 'https://demo.example.org:8010/', 'USE_SSL': True})
    8010

    """
    return (
        urlparse(settings_dict["SERVER_BASE_URL"]).port
        or (settings_dict["USE_SSL"] and 443)
        or 80
    )


url_parse_server_port.required_settings = ["SERVER_BASE_URL", "USE_SSL"]


def url_parse_server_protocol(settings_dict):
    """Return the public HTTP protocol, given the public base URL

    >>> url_parse_server_protocol({'USE_SSL': True})
    'https'

    >>> url_parse_server_protocol({'USE_SSL': False})
    'http'

    """
    return "https" if settings_dict["USE_SSL"] else "http"


url_parse_server_protocol.required_settings = ["USE_SSL"]


def url_parse_prefix(settings_dict):
    """Return the public URL prefix, given the public base URL

    >>> url_parse_prefix({'SERVER_BASE_URL': 'https://demo.example.org/demo/'})
    '/demo/'
    >>> url_parse_prefix({'SERVER_BASE_URL': 'http://demo.example.org/'})
    '/'
    >>> url_parse_prefix({'SERVER_BASE_URL': 'https://demo.example.org:8010'})
    '/'

    """
    p = urlparse(settings_dict["SERVER_BASE_URL"]).path
    if not p.endswith("/"):
        p += "/"
    return p


url_parse_prefix.required_settings = ["SERVER_BASE_URL"]


def url_parse_ssl(settings_dict):
    """Return True if the public URL uses https

    >>> url_parse_ssl({'SERVER_BASE_URL': 'https://demo.example.org/demo/'})
    True
    >>> url_parse_ssl({'SERVER_BASE_URL': 'http://demo.example.org/'})
    False

    """
    return urlparse(settings_dict["SERVER_BASE_URL"]).scheme == "https"


url_parse_ssl.required_settings = ["SERVER_BASE_URL"]


def use_x_forwarded_for(settings_dict):
    """Return `True` if this server is assumed to be behind a reverse proxy.
     Heuristic: the external port (in SERVER_PORT) is different from the actually listened port (in LISTEN_ADDRESS).

     >>> use_x_forwarded_for({'SERVER_PORT': 8000, 'LISTEN_ADDRESS': 'localhost:8000'})
     False
     >>> use_x_forwarded_for({'SERVER_PORT': 443, 'LISTEN_ADDRESS': 'localhost:8000'})
     True

    """
    listen_address, sep, listen_port = settings_dict["LISTEN_ADDRESS"].rpartition(":")
    if not re.match(r"\d+", listen_port):
        raise ValueError("Invalid LISTEN_ADDRESS port %s" % listen_port)
    return int(listen_port) != settings_dict["SERVER_PORT"]


use_x_forwarded_for.required_settings = ["SERVER_PORT", "LISTEN_ADDRESS"]


def project_name(settings_dict):
    """Transform the base module name into a nicer project name

    >>> project_name({'DF_MODULE_NAME': 'my_project'})
    'My Project'

    :param settings_dict:
    :return:
    """

    return " ".join(
        [
            x.capitalize()
            for x in settings_dict["DF_MODULE_NAME"].replace("_", " ").split()
        ]
    )


project_name.required_settings = ["DF_MODULE_NAME"]


def generate_secret_key(django_ready, length=60):
    if not django_ready:
        return get_random_string(length=length)
    from django.conf import settings

    return settings.SECRET_KEY


def required_packages(settings_dict):
    """
    Return a sorted list of the Python packages required by the current project (with their dependencies).
    A warning is added for each missing package.

    :param settings_dict:
    :return:
    """

    def get_requirements(package_name, parent=None):
        try:
            yield str(package_name)
            d = get_distribution(package_name)
            for r in d.requires():
                for required_package in get_requirements(r, parent=package_name):
                    yield str(required_package)
        except DistributionNotFound:
            settings_check_results.append(
                missing_package(str(package_name), " by %s" % parent)
            )
        except VersionConflict:
            settings_check_results.append(
                missing_package(str(package_name), " by %s" % parent)
            )

    return list(
        sorted(
            set(
                get_requirements(
                    settings_dict["DF_MODULE_NAME"],
                    parent=settings_dict["DF_MODULE_NAME"],
                )
            )
        )
    )


required_packages.required_settings = ["DF_MODULE_NAME"]


class ExcludedDjangoCommands:
    required_settings = ["DEVELOPMENT", "USE_CELERY", "DEBUG"]

    def __call__(self, settings_dict):
        result = {"startproject", "diffsettings"}
        if not settings_dict["DEVELOPMENT"]:
            result |= {
                "startapp",
                "findstatic",
                "npm",
                "makemigrations",
                "makemessages",
                "inspectdb",
                "compilemessages",
                "remove_stale_contenttypes",
                "squashmigrations",
            }
        if not settings_dict["USE_CELERY"]:
            result |= {"celery", "worker"}
        if not settings_dict["DEBUG"] and not settings_dict["DEVELOPMENT"]:
            result |= {"testserver", "test", "runserver"}
        return result

    def __repr__(self):
        return "%s.%s" % (self.__module__, "excluded_django_commands")


excluded_django_commands = ExcludedDjangoCommands()



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <github@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################

SECRET_KEY = "-"
USE_X_SEND_FILE = False
X_ACCEL_REDIRECT = False


# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <github@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################

import importlib
import os
from collections import OrderedDict
from configparser import RawConfigParser
from typing import Union

SOCIAL_PROVIDER_APPS = {
    "allauth.socialaccount.providers.%s" % x
    for x in {
        "amazon",
        "angellist",
        "asana",
        "auth0",
        "baidu",
        "basecamp",
        "bitbucket",
        "bitbucket_oauth2",
        "bitly",
        "coinbase",
        "digitalocean",
        "discord",
        "douban",
        "draugiem",
        "edmodo",
        "eveonline",
        "evernote",
        "facebook",
        "feedly",
        "flickr",
        "foursquare",
        "fxa",
        "github",
        "gitlab",
        "google",
        "hubic",
        "instagram",
        "kakao",
        "line",
        "linkedin",
        "linkedin_oauth2",
        "mailru",
        "mailchimp",
        "naver",
        "odnoklassniki",
        "openid",
        "orcid",
        "paypal",
        "persona",
        "pinterest",
        "reddit",
        "robinhood",
        "shopify",
        "slack",
        "soundcloud",
        "spotify",
        "stackexchange",
        "stripe",
        "tumblr",
        "twentythreeandme",
        "twitch",
        "twitter",
        "untappd",
        "vimeo",
        "vk",
        "weibo",
        "weixin",
        "windowslive",
        "xing",
    }
}


# Full set of social providers


class SocialProviderConfiguration:
    help = (
        "Please read https://django-allauth.readthedocs.io/en/latest/providers.html#{{ provider_id }}\n"
        "Contact {{ provider_name }} to get authentication secrets. "
        "Callback URL should be {{ SERVER_BASE_URL }}accounts/{{ provider_id }}/login/callback/"
    )
    attributes = {
        "client_id": "App ID, or consumer key",
        "secret": "API secret, client secret, or consumer secret",
        "key": "Key (often optional)",
    }
    name_prefix = "df-"

    def __init__(
        self,
        provider_id: str,
        provider_name: str,
        provider_app: str,
        values: Union[dict, None] = None,
    ):
        self.provider_id = provider_id
        self.provider_name = provider_name
        self.provider_app = provider_app
        self.values = values

    def __str__(self):
        return self.provider_name

    @property
    def name(self):
        return "%s%s" % (self.name_prefix, self.provider_name)

    def query_kwargs(self):
        return {"name": self.name, "provider": self.provider_id}

    @property
    def help_text(self):
        from df_config.config.base import merger
        from django.template import Template, Context

        context = {}
        context.update(merger.settings)
        context.update(self.__dict__)
        return Template(self.help).render(Context(dict_=context))


def get_available_configurations() -> dict:
    """return a dict of all existing social account provider configurations"""
    existing_providers = {}
    available_configurations = {x.id: x for x in SOCIAL_PROVIDER_CONFIGURATIONS}
    for provider_app in SOCIAL_PROVIDER_APPS:
        try:
            provider_module = importlib.import_module("%s.provider" % provider_app)
            for provider_cls in getattr(provider_module, "provider_classes", []):
                config_cls = available_configurations.get(
                    provider_cls.id, SocialProviderConfiguration
                )
                config = config_cls(provider_cls.id, provider_cls.name, provider_app)
                existing_providers[provider_cls.id] = config
        except ImportError:
            pass
    return existing_providers


def get_loaded_configurations() -> OrderedDict:
    """return a list of configured social authentication backends from a config file
    given in settings.ALLAUTH_APPLICATIONS_CONFIG
    """
    from django.conf import settings

    parser = RawConfigParser()
    if os.path.isfile(settings.ALLAUTH_APPLICATIONS_CONFIG):
        parser.read([settings.ALLAUTH_APPLICATIONS_CONFIG])

    existing_providers = get_available_configurations()
    providers = OrderedDict()
    for section in parser.sections():
        if section not in existing_providers:
            continue
        provider_config = existing_providers[section]
        values = {
            key: parser.get(section, key)
            for key in parser.options(section)
            if key in provider_config.attributes
        }
        provider_config.values = values
        providers[provider_config.provider_id] = provider_config
    return providers


def migrate(read_only: bool = False) -> bool:
    """
    Create and configure database social apps.

    Configure backends only for the first Site object.

    :param read_only:
    :return: True if (read_only and a migrate is required) or (not read_only and modifications were performed)
    """
    # noinspection PyPackageRequirements,PyUnresolvedReferences
    from allauth.socialaccount.models import SocialApp
    from django.contrib.sites.models import Site

    expected_configurations = {}
    for config in get_loaded_configurations().values():
        expected_configurations[(config.name, config.provider_id)] = config
    action_required = False
    to_remove_db_app_ids = []
    to_create_db_apps = []
    db_apps = {}
    for social_app in SocialApp.objects.filter(
        name__startswith=SocialProviderConfiguration.name_prefix
    ):
        key = (social_app.name, social_app.provider)
        if key not in expected_configurations:
            to_remove_db_app_ids.append(social_app.pk)
            continue
        else:
            db_apps[key] = social_app
    for key, configuration in expected_configurations.items():
        if key not in db_apps:
            to_create_db_apps.append(
                SocialApp(
                    name=configuration.name,
                    provider=configuration.provider_id,
                    **configuration.values
                )
            )
            continue
        app = db_apps[key]
        save = False
        for k, v in configuration.values.items():
            if getattr(app, k) != v:
                setattr(app, k, v)
                save = True
        if save:
            action_required = True
            if not read_only:
                app.save()
    if to_remove_db_app_ids:
        action_required = True
        if not read_only:
            SocialApp.objects.filter(pk__in=to_remove_db_app_ids).delete()
    if to_create_db_apps:
        action_required = True
        if not read_only:
            SocialApp.objects.bulk_create(to_create_db_apps)

    db_site = Site.objects.filter(pk=1).first()
    if db_site:
        required_db_ids = {
            x[0]
            for x in SocialApp.objects.filter(
                name__startswith=SocialProviderConfiguration.name_prefix
            ).values_list("id")
        }
        q = SocialApp.sites.through
        existing_db_ids = {
            x[0]
            for x in q.objects.filter(
                site_id=db_site.pk, socialapp_id__in=required_db_ids
            ).values_list("socialapp_id")
        }
        to_create_db_through_app = [
            q(site_id=db_site.pk, socialapp_id=x)
            for x in required_db_ids
            if x not in existing_db_ids
        ]
        if to_create_db_through_app:
            action_required = True
            if not read_only:
                q.objects.bulk_create(to_create_db_through_app)
    return action_required


class GithubConfiguration(SocialProviderConfiguration):
    id = "github"
    help = """open https://github.com/settings/applications/new and enter:
    Application name: {{ DF_PROJECT_NAME }} ({{ SERVER_NAME }})
    Homepage URL: {{ SERVER_BASE_URL }}
    Application description: {{ DF_PROJECT_NAME }} (or anything else useful)
    Authorization callback URL: {{ SERVER_BASE_URL }}accounts/github/login/callback/
    
    After validation, you can copy the following values:
    """
    attributes = {"client_id": "Client ID", "secret": "Client Secret"}


class FacebookConfiguration(SocialProviderConfiguration):
    id = "facebook"
    help = """open https://developers.facebook.com/apps and create an app to obtain a key and secret key.
    After registration you will need to make it available to the public. 
    In order to do that your app first has to be reviewed by Facebook.

    Leave your App Domains empty and put {{ SERVER_BASE_URL }} in the section labeled Website with Facebook Login. 
    Note that you’ll need to add your site’s actual domain to this section once it goes live.

    After validation, you can copy the following values:
    """
    attributes = {"client_id": "Client ID", "secret": "Client Secret"}


SOCIAL_PROVIDER_CONFIGURATIONS = [GithubConfiguration]



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <github@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
import os
from collections import OrderedDict
from configparser import RawConfigParser

from django.core.checks import Error
from pkg_resources import DistributionNotFound, get_distribution

from df_config.checks import missing_package, settings_check_results
from df_config.config.dynamic_settings import ExpandIterable
from df_config.guesses.social_providers import SOCIAL_PROVIDER_APPS
from df_config.utils import is_package_present


def allauth_provider_apps(settings_dict):
    parser = RawConfigParser()
    config = settings_dict["ALLAUTH_APPLICATIONS_CONFIG"]
    if not os.path.isfile(config):
        return []
    # noinspection PyBroadException
    try:
        parser.read([config])
    except Exception:
        settings_check_results.append(
            Error("Invalid config file. %s" % config, obj="configuration")
        )
        return []
    return [
        parser.get(section, "django_app")
        for section in parser.sections()
        if parser.has_option(section, "django_app")
    ]


allauth_provider_apps.required_settings = ["ALLAUTH_APPLICATIONS_CONFIG"]


class InstalledApps:
    """Provide a complete `INSTALLED_APPS` list, transparently adding common third-party packages.
     Specifically handle apps required by django-allauth (one by allowed method).

    """

    default_apps = [
        ExpandIterable("DF_INSTALLED_APPS"),
    ]
    base_django_apps = [
        "df_config",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.messages",
        "django.contrib.humanize",
        "django.contrib.sitemaps",
        "django.contrib.sites",
        "django.contrib.staticfiles",
        "django.contrib.admin",
    ]
    common_third_parties = OrderedDict(
        [
            ("USE_WEBSOCKETS", ["df_websockets", "channels"]),
            (
                "USE_SITE",
                [
                    "df_site",
                    "bootstrap4",
                    "channels",
                    "dal",
                    "dal_select2",
                    "fontawesome_5",
                    "smart_selects",
                ],
            ),
            ("USE_DEBUG_TOOLBAR", ["debug_toolbar"]),
            ("USE_PIPELINE", ["pipeline"]),
            ("USE_PAM_AUTHENTICATION", ["django_pam"]),
            ("RAVEN_DSN", ["raven.contrib.django.raven_compat"]),
        ]
    )
    required_settings = [
        "DF_INSTALLED_APPS",
        "ALLAUTH_PROVIDER_APPS",
        "USE_ALL_AUTH",
    ] + list(common_third_parties)
    social_apps = SOCIAL_PROVIDER_APPS

    def __call__(self, settings_dict):
        apps = self.default_apps
        apps += self.process_django_allauth(settings_dict)
        apps += self.process_third_parties(settings_dict)
        apps += self.base_django_apps
        return apps

    def process_third_parties(self, settings_dict):
        result = []
        for k, v in self.common_third_parties.items():
            package_name = v[0].partition(".")[0]
            if not settings_dict[k]:
                continue
            elif not is_package_present(package_name):
                settings_check_results.append(missing_package(package_name, ""))
                continue
            result += v
        return result

    def process_django_allauth(self, settings_dict):
        if (
            not settings_dict["USE_ALL_AUTH"]
            and not settings_dict["ALLAUTH_PROVIDER_APPS"]
        ):
            return []
        try:
            get_distribution("django-allauth")
        except DistributionNotFound:
            settings_check_results.append(
                missing_package(
                    "django-allauth", " to use OAuth2 or OpenID authentication"
                )
            )
            return []
        if "django.contrib.sites" not in self.default_apps:
            settings_check_results.append(
                Error(
                    '"django.contrib.sites" app must be enabled.',
                    obj="configuration",
                    id="df_config.E001",
                )
            )
            return []
        result = [
            "allauth",
            "allauth.account",
            "allauth.socialaccount",
            "allauth.socialaccount.providers.openid",
        ]
        if settings_dict["ALLAUTH_PROVIDER_APPS"]:
            result += [
                k
                for k in settings_dict["ALLAUTH_PROVIDER_APPS"]
                if k in self.social_apps
            ]
        return result

    def __repr__(self):
        return "%s.%s" % (self.__module__, "installed_apps")


installed_apps = InstalledApps()


class Middlewares:
    use_cache_middleware = True
    base_django_middlewares = [
        "django.contrib.sessions.middleware.SessionMiddleware",
        "django.middleware.common.CommonMiddleware",
        "django.middleware.csrf.CsrfViewMiddleware",
        "django.contrib.auth.middleware.AuthenticationMiddleware",
        "django.middleware.locale.LocaleMiddleware",
        "django.contrib.messages.middleware.MessageMiddleware",
        "django.middleware.clickjacking.XFrameOptionsMiddleware",
        "django.middleware.security.SecurityMiddleware",
        "df_config.apps.middleware.DFConfigMiddleware",
    ]
    common_third_parties = OrderedDict(
        [
            ("USE_WEBSOCKETS", "df_websockets.middleware.WebsocketMiddleware"),
            ("USE_DEBUG_TOOLBAR", "debug_toolbar.middleware.DebugToolbarMiddleware"),
        ]
    )
    required_settings = ["DF_MIDDLEWARE"] + list(common_third_parties)
    social_apps = SOCIAL_PROVIDER_APPS

    def __call__(self, settings_dict):
        mw_list = []
        mw_list += self.base_django_middlewares
        mw_list.append(ExpandIterable("DF_MIDDLEWARE"))
        mw_list += self.process_third_parties(settings_dict)

        if self.use_cache_middleware:
            mw_list.insert(0, "django.middleware.cache.UpdateCacheMiddleware")
            mw_list.append("django.middleware.cache.FetchFromCacheMiddleware")
        return mw_list

    def process_third_parties(self, settings_dict):
        result = []
        for k, v in self.common_third_parties.items():
            package_name = v.partition(".")[0]
            if not settings_dict[k]:
                continue
            elif not is_package_present(package_name):
                settings_check_results.append(missing_package(package_name, ""))
                continue
            result.append(v)
        return result

    def __repr__(self):
        return "%s.%s" % (self.__module__, "middlewares")


middlewares = Middlewares()



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <github@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
import os
import tempfile
from unittest import TestCase

import pkg_resources
from django.http import HttpRequest
from django.utils.http import http_date

from df_config.utils import RangedChunkReader, ensure_dir, is_package_present, send_file


class PatchSettings:
    """
    Temporarily change some settings, and restore them when the context is exited.

    """

    def __init__(self, **kwargs):
        self.patched_settings = kwargs
        self.original_settings = {}

    def __enter__(self):
        from django.conf import settings

        for k, v in self.patched_settings.items():
            if hasattr(settings, k):
                self.original_settings[k] = getattr(settings, k)
            setattr(settings, k, v)

    def __exit__(self, exc_type, exc_val, exc_tb):
        from django.conf import settings

        for k in self.patched_settings:
            if k in self.original_settings:
                setattr(settings, k, self.original_settings[k])
            else:
                delattr(settings, k)


class TestIsPackagePresent(TestCase):
    def test_is_package_present(self):
        self.assertTrue(is_package_present("df_config"))
        self.assertFalse(is_package_present("flask2"))


class TestEnsureDir(TestCase):
    def test_ensure_dir(self):
        with tempfile.TemporaryDirectory() as dirname:
            ensure_dir("%s/parent1/dirname" % dirname, parent=False)
            self.assertTrue(os.path.isdir("%s/parent1/dirname" % dirname))
            ensure_dir("%s/parent1/dirname" % dirname)
            self.assertTrue(os.path.isdir("%s/parent1/dirname" % dirname))
            ensure_dir("%s/parent2/filename" % dirname, parent=True)
            self.assertTrue(os.path.isdir("%s/parent2" % dirname))
            self.assertFalse(os.path.isdir("%s/parent2/filename" % dirname))


class TestSendFile(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        os.environ.setdefault("DJANGO_SETTINGS_MODULE", "test_df_config.data.settings")

    def test_rangedfilereader(self):
        filename = pkg_resources.resource_filename(
            "test_df_config", "data/range_data.txt"
        )
        with open(filename, "rb") as fd:
            reader = RangedChunkReader(fd, [(0, 9)], chunk_size=5)
            content = list(reader)
        self.assertEqual([b"11111", b"1111\n"], content)
        with open(filename, "rb") as fd:
            reader = RangedChunkReader(fd, [(0, 9), (20, 29)], chunk_size=5)
            content = list(reader)
        self.assertEqual([b"11111", b"1111\n", b"33333", b"3333\n"], content)

    def test_range(self):
        request = HttpRequest()
        filename = pkg_resources.resource_filename(
            "test_df_config", "data/range_data.txt"
        )
        request.META = {"HTTP_RANGE": "bytes=20-29"}
        r = send_file(
            request,
            filename,
            mimetype="text/plain",
            attachment_filename="range_data.txt",
        )
        content = r.getvalue()
        self.assertEqual(b"333333333\n", content)

        expected_headers = [
            ("Content-Type", "text/plain"),
            ("Content-Range", "bytes 20-29/90"),
            ("Content-Length", "10"),
            ('Last-Modified', http_date(os.stat(filename).st_mtime)),
            ("Content-Disposition", 'inline; filename="range_data.txt"'),
        ]
        self.assertEqual(expected_headers, list(r.items()))
        self.assertEqual(206, r.status_code)
        r.close()

    def test_range_multiple(self):
        request = HttpRequest()
        filename = pkg_resources.resource_filename(
            "test_df_config", "data/range_data.txt"
        )
        request.META = {"HTTP_RANGE": "bytes=0-9, 20-29, 40-49"}
        r = send_file(
            request,
            filename,
            mimetype="text/plain",
            attachment_filename="range_data.txt",
        )
        content = r.getvalue()
        self.assertEqual(b"111111111\n333333333\n555555555\n", content)
        expected_headers = [
            ("Content-Type", "text/plain"),
            ("Content-Length", "30"),
            ('Last-Modified', http_date(os.stat(filename).st_mtime)),
            ("Content-Disposition", 'inline; filename="range_data.txt"'),
        ]
        self.assertEqual(expected_headers, list(r.items()))
        self.assertEqual(200, r.status_code)
        r.close()

    def test_no_range(self):
        request = HttpRequest()
        filename = pkg_resources.resource_filename(
            "test_df_config", "data/range_data.txt"
        )
        request.META = {}
        r = send_file(
            request,
            filename,
            mimetype="text/plain",
            attachment_filename="range_data.txt",
        )
        content = r.getvalue()
        self.assertEqual(
            b"111111111\n222222222\n333333333\n444444444\n555555555\n666666666\n777777777\n888888888\n999999999\n",
            content,
        )
        expected_headers = [
            ("Content-Type", "text/plain"),
            ("Content-Length", "90"),
            ('Last-Modified', http_date(os.stat(filename).st_mtime)),
            ("Content-Disposition", 'inline; filename="range_data.txt"'),
        ]
        self.assertEqual(expected_headers, list(r.items()))
        self.assertEqual(200, r.status_code)
        r.close()



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <github@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################

"""Default values for all Django settings
======================================

Define settings for deploying most of df_config-based websites or for running them in `DEBUG` mode.
Most of them are used by Django, some of them by common third-party packages and the other ones are
used by DjangoFloor.

df_config also allows references between settings: for example, you only defines `SERVER_BASE_URL`
(like 'https://www.example.com/site/' ) and `SERVER_NAME` ('www.example.com'), `SERVER_PORT` ('443'),
`USE_SSL` ('True'), `SERVER_PROTOCOL` ('https') and `URL_PREFIX` ('/site/') are deduced.

These settings are defined in :mod:`df_config.config.defaults`.
Settings that should be customized on each installation (like the server name or the database password) can be
written in .ini files. The mapping between the Python setting and the [section/option] system is defined in
:mod:`df_config.iniconf`.

.. literalinclude:: ../../../../../df_config/config/defaults.py
   :language: python
   :lines: 41-1000

"""
import os

from django.utils.translation import gettext_lazy as _

# ######################################################################################################################
#
# detect if some external packages are available, to automatically customize some settings
#
# ######################################################################################################################
from df_config.config.dynamic_settings import (
    AutocreateFile,
    AutocreateFileContent,
    CallableSetting,
    Directory,
    ExpandIterable,
    Path,
    SettingReference,
)
from df_config.guesses.apps import allauth_provider_apps, installed_apps, middlewares
from df_config.guesses.auth import (
    authentication_backends,
    ldap_attribute_map,
    ldap_boolean_attribute_map,
    ldap_group_class,
    ldap_group_search,
    ldap_user_search,
)
from df_config.guesses.databases import (
    cache_redis_url,
    cache_setting,
    celery_redis_url,
    databases,
    session_redis_dict,
    websocket_redis_dict,
)
from df_config.guesses.log import log_configuration
from df_config.guesses.misc import (
    DefaultListenAddress,
    allowed_hosts,
    excluded_django_commands,
    generate_secret_key,
    project_name,
    required_packages,
    secure_hsts_seconds,
    smart_hostname,
    template_setting,
    url_parse_prefix,
    url_parse_server_name,
    url_parse_server_port,
    url_parse_server_protocol,
    url_parse_ssl,
    use_x_forwarded_for,
)
from df_config.guesses.pipeline import (
    pipeline_compilers,
    pipeline_css_compressor,
    pipeline_js_compressor,
)
from df_config.guesses.staticfiles import (
    pipeline_enabled,
    static_finder,
    static_storage,
)
from df_config.utils import guess_version, is_package_present

try:
    import django_redis  # does not work with is_package_present (???)

    USE_REDIS_CACHE = True
except ImportError:
    django_redis = None
    USE_REDIS_CACHE = False
USE_CELERY = is_package_present("celery")
USE_REDIS_SESSIONS = is_package_present("redis_sessions")
USE_PIPELINE = is_package_present("pipeline")
USE_DEBUG_TOOLBAR = is_package_present("debug_toolbar")
USE_ALL_AUTH = is_package_present("allauth")
USE_WEBSOCKETS = is_package_present("df_websockets")
USE_SITE = is_package_present("df_site")

# ######################################################################################################################
#
# settings that could be kept as-is for most projects
# of course, you can override them in your default settings
#
# ######################################################################################################################
ADMINS = (("admin", "{ADMIN_EMAIL}"),)
ALLOWED_HOSTS = CallableSetting(allowed_hosts)
CACHE_URL = CallableSetting(cache_redis_url)
CACHES = CallableSetting(cache_setting)
CSRF_COOKIE_DOMAIN = "{SERVER_NAME}"
CSRF_TRUSTED_ORIGINS = ["{SERVER_NAME}", "{SERVER_NAME}:{SERVER_PORT}"]
DATABASES = CallableSetting(databases)

DEBUG = False
# you should create a "local_settings.py" with "DEBUG = True" at the root of your project
DEVELOPMENT = True
# display all commands (like "migrate" or "runserver") in manage.py
# if False, development-specific commands are hidden

DEFAULT_FROM_EMAIL = "webmaster@{SERVER_NAME}"
FILE_UPLOAD_TEMP_DIR = Directory("{LOCAL_PATH}/tmp-uploads")
INSTALLED_APPS = CallableSetting(installed_apps)
LOGGING = CallableSetting(log_configuration)
MANAGERS = SettingReference("ADMINS")
MEDIA_ROOT = Directory("{LOCAL_PATH}/media")
MEDIA_URL = "/media/"
MIDDLEWARE = CallableSetting(middlewares)

ROOT_URLCONF = "df_config.root_urls"
SECRET_KEY = AutocreateFileContent(
    "{LOCAL_PATH}/secret_key.txt", generate_secret_key, mode=0o600, length=60
)
SECURE_BROWSER_XSS_FILTER = True
SECURE_CONTENT_TYPE_NOSNIFF = True
SECURE_HSTS_INCLUDE_SUBDOMAINS = SettingReference("USE_SSL")
SECURE_HSTS_PRELOAD = SettingReference("USE_SSL")
SECURE_HSTS_SECONDS = CallableSetting(secure_hsts_seconds)
SECURE_PROXY_SSL_HEADER = (
    "HTTP_X_FORWARDED_PROTO",
    "https",
)  # X-Forwarded-Proto or None
SECURE_SSL_REDIRECT = SettingReference("USE_SSL")
SECURE_FRAME_DENY = SettingReference("USE_SSL")
SERVER_EMAIL = "{ADMIN_EMAIL}"
SESSION_COOKIE_AGE = 1209600
TEMPLATES = CallableSetting(template_setting)
TEMPLATE_DEBUG = SettingReference("DEBUG")
TEMPLATE_DIRS = ()
TEMPLATE_CONTEXT_PROCESSORS = [
    "django.contrib.auth.context_processors.auth",
    "django.template.context_processors.debug",
    "django.template.context_processors.i18n",
    "django.template.context_processors.media",
    "django.template.context_processors.static",
    "django.template.context_processors.tz",
    "django.contrib.messages.context_processors.messages",
    "df_config.context_processors.config",
    ExpandIterable("DF_TEMPLATE_CONTEXT_PROCESSORS"),
]
TEST_RUNNER = "django.test.runner.DiscoverRunner"
USE_I18N = True
USE_L10N = True
USE_THOUSAND_SEPARATOR = True
USE_TZ = True
USE_X_FORWARDED_HOST = True  # X-Forwarded-Host
X_FRAME_OPTIONS = "SAMEORIGIN"
WSGI_APPLICATION = "df_config.application.wsgi_application"

# django.contrib.auth
AUTHENTICATION_BACKENDS = CallableSetting(authentication_backends)
LOGIN_URL = "/admin/login/"
LOGIN_REDIRECT_URL = "{URL_PREFIX}"
# LOGOUT_REDIRECT_URL = '{URL_PREFIX}'

# django.contrib.sessions
if USE_REDIS_SESSIONS:
    SESSION_ENGINE = "redis_sessions.session"
SESSION_COOKIE_SECURE = SettingReference("USE_SSL")
CSRF_COOKIE_SECURE = SettingReference("USE_SSL")

# django.contrib.sites
SITE_ID = 1

# django.contrib.staticfiles
STATIC_ROOT = Directory("{LOCAL_PATH}/static")
STATIC_URL = "/static/"
STATICFILES_STORAGE = CallableSetting(static_storage)
STATICFILES_FINDERS = CallableSetting(static_finder)

# celery
BROKER_URL = CallableSetting(celery_redis_url)
CELERY_DEFAULT_QUEUE = "celery"
CELERY_TIMEZONE = "{TIME_ZONE}"
CELERY_RESULT_EXCHANGE = "{DF_MODULE_NAME}_results"
CELERY_RESULT_BACKEND = "{BROKER_URL}"
CELERY_RESULT_SERIALIZER = "json"
CELERY_ACCEPT_CONTENT = ["json", "yaml", "msgpack"]
CELERY_APP = "df_websockets"
CELERY_CREATE_DIRS = True
CELERY_TASK_SERIALIZER = "json"

# df_config
DF_PROJECT_VERSION = CallableSetting(guess_version)
DATA_PATH = Directory("{LOCAL_PATH}/data")
SERVER_NAME = CallableSetting(url_parse_server_name)  # ~ www.example.org
SERVER_PORT = CallableSetting(url_parse_server_port)  # ~ 443
SERVER_PROTOCOL = CallableSetting(url_parse_server_protocol)  # ~ "https"
URL_PREFIX = CallableSetting(url_parse_prefix)
# something like "/prefix/" (but probably just "/")
USE_SSL = CallableSetting(url_parse_ssl)  # ~ True
USE_X_SEND_FILE = False  # Apache module
X_ACCEL_REDIRECT = []  # paths used by nginx
USE_HTTP_BASIC_AUTH = False  # HTTP-Authorization
USE_X_FORWARDED_FOR = CallableSetting(use_x_forwarded_for)  # X-Forwarded-For
DF_FAKE_AUTHENTICATION_USERNAME = None
DF_ALLOW_USER_CREATION = True

DF_SERVER = "gunicorn"  # "gunicorn" or "daphne"
DF_REMOVED_DJANGO_COMMANDS = CallableSetting(excluded_django_commands)
DF_ALLOW_LOCAL_USERS = True
DF_CHECKED_REQUIREMENTS = CallableSetting(required_packages)

# df_websockets
WEBSOCKET_URL = "/ws/"  # set to None if you do not use websockets
WEBSOCKET_REDIS_CONNECTION = CallableSetting(websocket_redis_dict)
WEBSOCKET_SIGNAL_DECODER = "json.JSONDecoder"
WEBSOCKET_TOPIC_SERIALIZER = "df_websockets.topics.serialize_topic"
WEBSOCKET_SIGNAL_ENCODER = "django.core.serializers.json.DjangoJSONEncoder"
WEBSOCKET_REDIS_PREFIX = "ws"
WEBSOCKET_REDIS_EXPIRE = 36000
WINDOW_INFO_MIDDLEWARES = [
    "df_websockets.ws_middleware.WindowKeyMiddleware",
    "df_websockets.ws_middleware.DjangoAuthMiddleware",
    "df_websockets.ws_middleware.Djangoi18nMiddleware",
    "df_websockets.ws_middleware.BrowserMiddleware",
]
if USE_WEBSOCKETS:
    ASGI_APPLICATION = "df_websockets.routing.application"
else:
    ASGI_APPLICATION = "df_config.application.asgi_application"

# django-channels
CHANNEL_LAYERS = {
    "default": {
        "BACKEND": "channels_redis.core.RedisChannelLayer",
        "CONFIG": {
            "hosts": [
                {
                    "address": (
                        "{WEBSOCKET_REDIS_HOST}",
                        SettingReference("WEBSOCKET_REDIS_PORT"),
                    ),
                    "password": SettingReference("WEBSOCKET_REDIS_PASSWORD"),
                    "db": SettingReference("WEBSOCKET_REDIS_DB"),
                }
            ]
        },
    }
}

# django-pipeline
PIPELINE = {
    "PIPELINE_ENABLED": SettingReference("PIPELINE_ENABLED"),
    "JAVASCRIPT": SettingReference("PIPELINE_JS"),
    "STYLESHEETS": SettingReference("PIPELINE_CSS"),
    "CSS_COMPRESSOR": SettingReference("PIPELINE_CSS_COMPRESSOR"),
    "JS_COMPRESSOR": SettingReference("PIPELINE_JS_COMPRESSOR"),
    "COMPILERS": SettingReference("PIPELINE_COMPILERS"),
}
PIPELINE_COMPILERS = CallableSetting(pipeline_compilers)
PIPELINE_CSS_COMPRESSOR = CallableSetting(pipeline_css_compressor)
PIPELINE_JS_COMPRESSOR = CallableSetting(pipeline_js_compressor)
PIPELINE_CSS = {
    "django.admin": {
        "source_filenames": ["admin/css/base.css", "admin/css/responsive.css"],
        "output_filename": "css/django-admin.css",
        "extra_context": {"media": "all"},
    },
    "default": {
        "source_filenames": [ExpandIterable("DF_CSS")],
        "output_filename": "css/default.css",
        "extra_context": {"media": "all"},
    },
}
PIPELINE_ENABLED = CallableSetting(pipeline_enabled)
PIPELINE_JS = {
    "django.admin": {"source_filenames": [], "output_filename": "js/django-admin.js"},
    "default": {
        "source_filenames": [ExpandIterable("DF_JS")],
        "output_filename": "js/default.js",
    },
}
LIVE_SCRIPT_BINARY = "lsc"
LESS_BINARY = "lessc"
SASS_BINARY = "sass"
STYLUS_BINARY = "stylus"
BABEL_BINARY = "babel"
YUGLIFY_BINARY = "yuglify"
YUI_BINARY = "yuicompressor"
CLOSURE_BINARY = "closure"
UGLIFYJS_BINARY = "uglifyjs"
CSSTIDY_BINARY = "csstidy"
COFFEE_SCRIPT_BINARY = "coffee"
CSSMIN_BINARY = "cssmin"
TYPESCRIPT_BINARY = "tsc"
TYPESCRIPT_ARGUMENTS = []
CSSNANO_BINARY = "cssnano"
CSSNANO_ARGUMENTS = []
TERSER_BINARY = "terser"
TERSER_ARGUMENTS = []

# Django-All-Auth
ACCOUNT_EMAIL_SUBJECT_PREFIX = "[{SERVER_NAME}] "
ACCOUNT_EMAIL_VERIFICATION = None
ALLAUTH_PROVIDER_APPS = CallableSetting(allauth_provider_apps)
ALLAUTH_APPLICATIONS_CONFIG = AutocreateFile("{LOCAL_PATH}/social_auth.ini", mode=0o600)
ACCOUNT_EMAIL_REQUIRED = True
ACCOUNT_AUTHENTICATION_METHOD = "username_email"
ACCOUNT_DEFAULT_HTTP_PROTOCOL = "{SERVER_PROTOCOL}"
ACCOUNT_ADAPTER = "df_config.apps.allauth.AccountAdapter"

# Django-Debug-Toolbar
DEBUG_TOOLBAR_CONFIG = {"JQUERY_URL": "{STATIC_URL}vendor/jquery/dist/jquery.min.js"}
DEBUG_TOOLBAR_PATCH_SETTINGS = False
DEBUG_TOOLBAR_PANELS = [
    "debug_toolbar.panels.versions.VersionsPanel",
    "debug_toolbar.panels.timer.TimerPanel",
    "debug_toolbar.panels.settings.SettingsPanel",
    "debug_toolbar.panels.profiling.ProfilingPanel",
    "debug_toolbar.panels.headers.HeadersPanel",
    "debug_toolbar.panels.request.RequestPanel",
    "debug_toolbar.panels.sql.SQLPanel",
    "debug_toolbar.panels.templates.TemplatesPanel",
    "debug_toolbar.panels.staticfiles.StaticFilesPanel",
    "debug_toolbar.panels.cache.CachePanel",
    "debug_toolbar.panels.signals.SignalsPanel",
    "debug_toolbar.panels.redirects.RedirectsPanel",
]
INTERNAL_IPS = ("127.0.0.1", "::1", "localhost")

# django-auth-ldap
AUTH_LDAP_SERVER_URI = None
AUTH_LDAP_BIND_DN = ""
AUTH_LDAP_BIND_PASSWORD = ""
AUTH_LDAP_USER_SEARCH_BASE = "ou=users,dc=example,dc=com"
AUTH_LDAP_FILTER = "(uid=%(user)s)"
AUTH_LDAP_USER_SEARCH = CallableSetting(ldap_user_search)
AUTH_LDAP_USER_DN_TEMPLATE = None
AUTH_LDAP_START_TLS = False
AUTH_LDAP_USER_ATTR_MAP = CallableSetting(ldap_attribute_map)
AUTH_LDAP_USER_FLAGS_BY_GROUP = CallableSetting(ldap_boolean_attribute_map)
AUTH_LDAP_MIRROR_GROUPS = False
AUTH_LDAP_USER_IS_ACTIVE = None
AUTH_LDAP_USER_IS_STAFF = None
AUTH_LDAP_USER_IS_SUPERUSER = None
AUTH_LDAP_USER_FIRST_NAME = None
AUTH_LDAP_USER_LAST_NAME = None
AUTH_LDAP_USER_EMAIL = None
AUTH_LDAP_GROUP_TYPE = CallableSetting(ldap_group_class)
AUTH_LDAP_GROUP_NAME = "posix"
AUTH_LDAP_ALWAYS_UPDATE_USER = True
AUTH_LDAP_REQUIRE_GROUP = None
AUTH_LDAP_DENY_GROUP = None
# Cache group memberships for an hour to minimize LDAP traffic
AUTH_LDAP_CACHE_GROUPS = True
AUTH_LDAP_GROUP_CACHE_TIMEOUT = 3600
# Use LDAP group membership to calculate group permissions.
AUTH_LDAP_FIND_GROUP_PERMS = False
AUTH_LDAP_GROUP_SEARCH = CallableSetting(ldap_group_search)
AUTH_LDAP_GROUP_SEARCH_BASE = "ou=groups,dc=example,dc=com"
AUTH_LDAP_AUTHORIZE_ALL_USERS = True
# https://bitbucket.org/illocution/django-auth-ldap/pull-requests/29/kerberos-bind-method-to-provide-multi/diff
# KRB5_CCACHE = None
# KRB5_KEYTAB = None
# KRB5_PRINCIPAL = None

# django-cors-headers
CORS_ORIGIN_WHITELIST = ("{SERVER_NAME}", "{SERVER_NAME}:{SERVER_PORT}")
CORS_REPLACE_HTTPS_REFERER = False

# django-hosts
DEFAULT_HOST = "{SERVER_NAME}"
HOST_SCHEME = "{SERVER_PROTOCOL}://"
HOST_PORT = "{SERVER_PORT}"

# django—pam
USE_PAM_AUTHENTICATION = False

# django-radius
RADIUS_SERVER = None
RADIUS_PORT = None
RADIUS_SECRET = None

# django-redis-sessions
SESSION_REDIS = CallableSetting(session_redis_dict)

# django-smart-selects
USE_DJANGO_JQUERY = True
JQUERY_URL = False

# raven
RAVEN_CONFIG = {"dsn": "{RAVEN_DSN}", "release": "{DF_PROJECT_VERSION}"}

# ######################################################################################################################
#
# settings that should be customized for each project
# of course, you can redefine or override any setting
#
# ######################################################################################################################
# df_config
DF_CSS = []
DF_JS = []
DF_INDEX_VIEW = None
DF_SITE_SEARCH_VIEW = None  # 'djangofloor.views.search.UserSearchView'
DF_PROJECT_NAME = CallableSetting(project_name)
DF_URL_CONF = "{DF_MODULE_NAME}.urls.urlpatterns"
DF_ADMIN_SITE = "django.contrib.admin.site"
DF_JS_CATALOG_VIEWS = ["django.contrib.admin"]
# noinspection PyUnresolvedReferences
DF_INSTALLED_APPS = ["{DF_MODULE_NAME}"]  # your django app!
DF_MIDDLEWARE = []
DF_TEMPLATE_CONTEXT_PROCESSORS = []
DF_PIP_NAME = "{DF_MODULE_NAME}"  # anything such that "pip install {DF_PIP_NAME}" installs your project
# only used in docs
DF_REMOTE_USER_HEADER = None  # HTTP_REMOTE_USER
DF_DEFAULT_GROUPS = [_("Users")]
NPM_FILE_PATTERNS = {
    "bootstrap-notify": ["*.js"],
    "font-awesome": ["css/*", "fonts/*"],
    "html5shiv": ["dist/*"],
    "jquery": ["dist/*"],
    "jquery-file-upload": ["css/*", "js/*"],
    "respond.js": ["dest/*"],
}
# used by the "npm" command: downloads these packages and copies the files matching any pattern in the list
LOG_REMOTE_ACCESS = True
LOG_DIRECTORY = Directory("{LOCAL_PATH}/log")
LOG_EXCLUDED_COMMANDS = {
    "clearsessions",
    "check",
    "compilemessages",
    "collectstatic",
    "config",
    "createcachetable",
    "changepassword",
    "createsuperuser",
    "dumpdb",
    "dbshell",
    "dumpdata",
    "flush",
    "loaddata",
    "inspectdb",
    "makemessages",
    "makemigrations",
    "migrate",
    "npm",
    "packaging",
    "ping_google",
    "remove_stale_contenttypes",
    "sendtestemail",
    "shell",
    "showmigrations",
    "sqlflush",
    "sqlmigrate",
    "sqlsequencereset",
    "squashmigrations",
    "startapp",
    "test",
    "testserver",
}

# ######################################################################################################################
#
# settings that should be customized for each deployment
# {DF_MODULE_NAME}.iniconf:INI_MAPPING should be a list of ConfigField, allowing to define these settings in a .ini file
#
# ######################################################################################################################
ADMIN_EMAIL = "admin@{SERVER_NAME}"  # aliased in settings.ini as "[global]admin_email"
DATABASE_ENGINE = "sqlite3"  # aliased in settings.ini as "[database]engine"
DATABASE_NAME = Path(
    "{LOCAL_PATH}/database.sqlite3"
)  # aliased in settings.ini as "[database]name"
DATABASE_USER = ""  # aliased in settings.ini as "[database]user"
DATABASE_PASSWORD = ""  # aliased in settings.ini as "[database]password"
DATABASE_HOST = ""  # aliased in settings.ini as "[database]host"
DATABASE_PORT = ""  # aliased in settings.ini as "[database]port"
DATABASE_OPTIONS = {}
EMAIL_HOST = "localhost"  # aliased in settings.ini as "[email]host"
EMAIL_HOST_PASSWORD = ""  # aliased in settings.ini as "[email]password"
EMAIL_HOST_USER = ""  # aliased in settings.ini as "[email]user"
EMAIL_FROM = "{ADMIN_EMAIL}"  # aliased in settings.ini as "[email]from"
EMAIL_PORT = 25  # aliased in settings.ini as "[email]port"
EMAIL_SUBJECT_PREFIX = "[{SERVER_NAME}] "
EMAIL_USE_TLS = False  # aliased in settings.ini as "[email]use_tls"
EMAIL_USE_SSL = False  # aliased in settings.ini as "[email]use_ssl"
EMAIL_SSL_CERTFILE = None
EMAIL_SSL_KEYFILE = None
LANGUAGE_CODE = "en"  # aliased in settings.ini as "[global]language_code"
TIME_ZONE = "Europe/Paris"  # aliased in settings.ini as "[global]time_zone"
LOG_REMOTE_URL = None  # aliased in settings.ini as "[global]log_remote_url"
LOG_LEVEL = None
SERVER_BASE_URL = CallableSetting(
    smart_hostname
)  # aliased in settings.ini as "[global]server_url"

# df_config
LISTEN_ADDRESS = DefaultListenAddress(
    9000
)  # aliased in settings.ini as "[global]listen_address"
LOCAL_PATH = "./django_data"  # aliased in settings.ini as "[global]data"
__split_path = __file__.split(os.path.sep)
if "lib" in __split_path:
    prefix = os.path.join(*__split_path[: __split_path.index("lib")])
    LOCAL_PATH = Directory("/%s/var/{DF_MODULE_NAME}" % prefix)

# django-redis-sessions
SESSION_REDIS_PROTOCOL = "redis"
SESSION_REDIS_HOST = "localhost"  # aliased in settings.ini as "[session]host"
SESSION_REDIS_PORT = 6379  # aliased in settings.ini as "[session]port"
SESSION_REDIS_DB = 1  # aliased in settings.ini as "[session]db"
SESSION_REDIS_PASSWORD = ""  # aliased in settings.ini as "[session]password"

# django_redis (cache)
CACHE_PROTOCOL = "redis"
CACHE_HOST = "localhost"  # aliased in settings.ini as "[cache]host"
CACHE_PORT = 6379  # aliased in settings.ini as "[cache]port"
CACHE_DB = 2  # aliased in settings.ini as "[cache]db"
CACHE_PASSWORD = ""  # aliased in settings.ini as "[cache]password"

# websockets
WEBSOCKET_REDIS_PROTOCOL = "redis"
WEBSOCKET_REDIS_HOST = "localhost"  # aliased in settings.ini as "[websocket]host"
WEBSOCKET_REDIS_PORT = 6379  # aliased in settings.ini as "[websocket]port"
WEBSOCKET_REDIS_DB = 3  # aliased in settings.ini as "[websocket]db"
WEBSOCKET_REDIS_PASSWORD = None  # aliased in settings.ini as "[websocket]password"

# celery
CELERY_PROTOCOL = "redis"
CELERY_HOST = "localhost"  # aliased in settings.ini as "[celery]host"
CELERY_PORT = 6379  # aliased in settings.ini as "[celery]port"
CELERY_DB = 4  # aliased in settings.ini as "[celery]db"
CELERY_PASSWORD = ""  # aliased in settings.ini as "[celery]password"
CELERY_PROCESSES = 4

# raven
RAVEN_DSN = None



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <github@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
import logging

from django.conf import settings
from django.contrib.auth.backends import RemoteUserBackend
from django.contrib.auth.models import Group
from django.utils.functional import cached_property

logger = logging.getLogger("django.request")

_CACHED_GROUPS = {}


class DefaultGroupsRemoteUserBackend(RemoteUserBackend):
    """Add groups to new users.
    Based on :class:`django.contrib.auth.backends.RemoteUserBackend`.
    Only overrides the `configure_user` method to add the required groups.

     """

    @property
    def create_unknown_user(self):
        return settings.DF_ALLOW_USER_CREATION

    @cached_property
    def ldap_backend(self):
        # noinspection PyUnresolvedReferences
        from django_auth_ldap.backend import LDAPBackend

        return LDAPBackend()

    def authenticate(self, *args, **kwargs):
        remote_user = kwargs.get("remote_user")
        if (
            remote_user
            and settings.AUTH_LDAP_SERVER_URI
            and settings.AUTH_LDAP_ALWAYS_UPDATE_USER
        ):
            user = self.ldap_backend.populate_user(remote_user)
            if user:
                return user
        return super().authenticate(*args, remote_user)

    def configure_user(self, request, user=None):
        """
        Configures a user after creation and returns the updated user.

        By default, returns the user unmodified; only add it to the default group.
        """
        if user is None:  # compatibility
            user = request
        for group_name in settings.DF_DEFAULT_GROUPS:
            if group_name not in _CACHED_GROUPS:
                _CACHED_GROUPS[group_name] = Group.objects.get_or_create(
                    name=str(group_name)
                )[0]
            user.groups.add(_CACHED_GROUPS[group_name])
        return user



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <github@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################

"""Convert values from string to Python values
===========================================

Use these classes in your mapping provided in `yourproject.iniconf:INI_MAPPING`.
Check :mod:`df_config.iniconf` for examples.
"""

import os
from typing import Any

from django.core.checks import Error

from df_config.checks import settings_check_results

MISSING_VALUE = [[]]


def bool_setting(value):
    """return `True` if the provided (lower-cased) text is one of {'1', 'ok', 'yes', 'true', 'on'}"""
    return str(value).lower() in {"1", "ok", "yes", "true", "on"}


def str_or_none(text):
    """return `None` if the text is empty, else returns the text"""
    return text or None


def str_or_blank(value):
    """return '' if the provided value is `None`, else return value"""
    return "" if value is None else str(value)


def guess_relative_path(value):
    """Replace an absolute path by its relative path if the abspath begins by the current dir"""
    if not value:
        return ""
    value = os.path.abspath(value)
    cwd = os.getcwd()
    if value.startswith(cwd):
        return ".%s" % (value[len(cwd) :])
    return value


def strip_split(value):
    """Split the value on "," and strip spaces of the result. Remove empty values.

    >>> strip_split('keyword1, keyword2 ,,keyword3')
    ['keyword1', 'keyword2', 'keyword3']

    >>> strip_split('')
    []

    >>> strip_split(None)
    []

    :param value:
    :type value:
    :return: a list of strings
    :rtype: :class:`list`
    """
    if value:
        return [x.strip() for x in value.split(",") if x.strip()]
    return []


class ConfigField:
    """Class that maps an option in a .ini file to a setting.

    :param name: the section and the option in a .ini file (like "database.engine")
    :param setting_name: the name of the setting (like "DATABASE_ENGINE")
    :param from_str: any callable that takes a text value and returns an object. Default to `str_or_none`
    :type from_str: `callable`
    :param to_str: any callable that takes the Python value and that converts it to str. Default to `str`
    :type to_str: `callable`
    :param help_str: any text that can serve has help in documentation.
    :param default: the value that will be used in documentation. The current setting value is used if equal to `None`.
    """

    def __init__(
        self,
        name: str,
        setting_name: str,
        from_str=str,
        to_str=str_or_blank,
        help_str: str = None,
        default: Any = None,
    ):
        self.name = name
        self.setting_name = setting_name
        self.from_str = from_str
        self.to_str = to_str
        self.__doc__ = help_str
        self.value = default

    def __str__(self):
        return self.name


class CharConfigField(ConfigField):
    """Accepts str values. If `allow_none`, then `None` replaces any empty value."""

    def __init__(self, name, setting_name, allow_none=True, **kwargs):
        from_str = str_or_none if allow_none else str
        super().__init__(
            name, setting_name, from_str=from_str, to_str=str_or_blank, **kwargs
        )


class IntegerConfigField(ConfigField):
    """Accept integer values. If `allow_none`, then `None` replaces any empty values (other `0` is used)."""

    def __init__(self, name, setting_name, allow_none=True, **kwargs):
        if allow_none:

            def from_str(value: str):
                return int(value) if value else None

        else:

            def from_str(value: str):
                return int(value) if value else 0

        super().__init__(
            name, setting_name, from_str=from_str, to_str=str_or_blank, **kwargs
        )


class FloatConfigField(ConfigField):
    """Accept floating-point values. If `allow_none`, then `None` replaces any empty values (other `0.0` is used)."""

    def __init__(self, name, setting_name, allow_none=True, **kwargs):
        if allow_none:

            def from_str(value):
                return float(value) if value else None

        else:

            def from_str(value):
                return float(value) if value else 0.0

        super().__init__(
            name, setting_name, from_str=from_str, to_str=str_or_blank, **kwargs
        )


class ListConfigField(ConfigField):
    """Convert a string to a list of values, splitted with the :meth:`df_config.config.fields.strip_split` function."""

    def __init__(self, name, setting_name, **kwargs):
        def to_str(value):
            if value:
                return ",".join([str(x) for x in value])
            return ""

        super().__init__(
            name, setting_name, from_str=strip_split, to_str=to_str, **kwargs
        )


class BooleanConfigField(ConfigField):
    """Search for a boolean value in the ini file.
    If this value is empty and `allow_none` is `True`, then the value is `None`.
    Otherwise returns `True` if the provided (lower-cased) text is one of ('1', 'ok', 'yes', 'true', 'on')
    """

    def __init__(self, name, setting_name, allow_none=False, **kwargs):
        if allow_none:

            def from_str(value):
                if not value:
                    return None
                return bool_setting(value)

            def to_str(value):
                if value is None:
                    return ""
                return str(bool(value)).lower()

        else:

            def from_str(value):
                return bool_setting(value)

            def to_str(value):
                return str(bool(value)).lower()

        super().__init__(name, setting_name, from_str=from_str, to_str=to_str, **kwargs)


class ChoiceConfigFile(ConfigField):
    """Only allow a limited set of values in the .ini file.
    The available values must be given as :class:`str`.

    Choices must be a :class:`dict`, mapping .ini (string) values to actual values.

    If an invalid value is provided by the user, then `None` is returned, but an error is displayed
    through the Django check system.
    """

    def __init__(self, name, setting_name, choices, help_str="", **kwargs):
        def from_str(value):
            if value not in choices:
                valid = ", ".join(['"%s"' % x for x in choices])
                settings_check_results.append(
                    Error(
                        'Invalid value "%s". Valid choices: %s.' % (value, valid),
                        obj="configuration",
                    )
                )
            return choices.get(value)

        def to_str(value):
            for k, v in choices.items():
                if v == value:
                    return str(k)
            return ""

        valid_values = ", ".join(['"%s"' % x for x in choices])
        if help_str:
            help_str += " Valid choices: %s" % valid_values
        else:
            help_str = "Valid choices: %s" % valid_values

        super().__init__(
            name,
            setting_name,
            from_str=from_str,
            to_str=to_str,
            help_str=help_str,
            **kwargs
        )



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <github@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
import shutil
from typing import Dict, List, Optional, Tuple, Union

from df_config.utils import is_package_present

available_css_compressor = [
    ("df_config.apps.pipeline.CssNanoCompressor", "CSSNANO_BINARY", None),
    ("pipeline.compressors.yuglify.YuglifyCompressor", "YUGLIFY_BINARY", None),
    ("pipeline.compressors.yui.YUICompressor", "YUI_BINARY", None),
    ("pipeline.compressors.csstidy.CSSTidyCompressor", "CSSTIDY_BINARY", None),
    ("pipeline.compressors.cssmin.CSSMinCompressor", "CSSMIN_BINARY", None),
    ("df_config.apps.pipeline.RcssCompressor", None, "rcssmin"),
    ("pipeline.compressors.NoopCompressor", None, None),
]
available_js_compressors = [
    ("df_config.apps.pipeline.TerserCompressor", "TERSER_BINARY", None),
    ("pipeline.compressors.uglifyjs.UglifyJSCompressor", "UGLIFYJS_BINARY", None),
    ("pipeline.compressors.yuglify.YuglifyCompressor", "YUGLIFY_BINARY", None),
    ("pipeline.compressors.yui.YUICompressor", "YUI_BINARY", None),
    ("pipeline.compressors.closure.ClosureCompressor", "CLOSURE_BINARY", None),
    ("pipeline.compressors.jsmin.JSMinCompressor", None, "jsmin"),
    ("pipeline.compressors.slimit.SlimItCompressor", None, "slimit"),
    ("pipeline.compressors.NoopCompressor", None, None),
]
available_compilers = [
    ("pipeline.compilers.coffee.CoffeeScriptCompiler", "COFFEE_SCRIPT_BINARY", None),
    ("pipeline.compilers.livescript.LiveScriptCompiler", "LIVE_SCRIPT_BINARY", None),
    ("pipeline.compilers.less.LessCompiler", "LESS_BINARY", None),
    ("pipeline.compilers.sass.SASSCompiler", "SASS_BINARY", None),
    ("pipeline.compilers.stylus.StylusCompiler", "STYLUS_BINARY", None),
    ("pipeline.compilers.es6.ES6Compiler", "BABEL_BINARY", None),
    ("df_config.apps.pipeline.TypescriptCompiler", "TYPESCRIPT_BINARY", None),
    ("df_config.apps.pipeline.PyScssCompiler", None, "scss"),
]


def required_settings(
        candidates: List[Tuple[str, Optional[str], Optional[str]]]
) -> List[str]:
    return [x[1] for x in candidates if x[1]]


def guess_pipeline_extension(
        candidates: List[Tuple[str, Optional[str], Optional[str]]],
        settings_dict: Dict,
        one: bool = False,
) -> Union[List[str], Optional[str]]:
    extensions = []
    for extension, setting, package in candidates:
        if package and not is_package_present(package):
            continue
        if setting and not shutil.which(settings_dict[setting]):
            continue
        if one:
            return extension
        extensions.append(extension)
    if one:
        return None
    return extensions


def pipeline_css_compressor(settings_dict: Dict) -> str:
    return guess_pipeline_extension(available_css_compressor, settings_dict, one=True)


pipeline_css_compressor.required_settings = required_settings(available_css_compressor)


def pipeline_js_compressor(settings_dict: Dict) -> str:
    return guess_pipeline_extension(available_js_compressors, settings_dict, one=True)


pipeline_js_compressor.required_settings = required_settings(available_js_compressors)


def pipeline_compilers(settings_dict: Dict) -> List[str]:
    return guess_pipeline_extension(available_compilers, settings_dict)


pipeline_compilers.required_settings = required_settings(available_compilers)
